 body {
    font-family: sans-serif;
    font-size: 14px;
    margin: 0;
    padding-left: 30px;
    padding-right: 30px;
    background-color: #fafafa;
  }
  
  .vertical-section-container {
    height: 600;
    display: flex;
    flex-direction: column;
  }

  .centered {
    max-width: 100%;
    height: 600;
    margin-left: auto;
    margin-right: auto;
    border: none;
    position: relative;
    left: 0px;
  }

  .chart {
    position: absolute;
    top: 40px;
    left: 0px;   
  }

  .meta {
    position: absolute;
    top: 20px;
    left: 0px;   
  }

  .bottom {
    border-bottom: 1px solid #006080;
  }

  .full {
    width: 100%;
  }

  .labels {
    font-family: "Arial";
    font-weight: bold;
    font-size: 20px;
  }

  .labels_small {
    font-family: "Arial";
    font-weight: bold;
    font-size: 16px;
  }

  .header {
    font-family: "Arial";
    font-weight: bold;
    font-size: 24px;
    padding-left: 80px;
  }

  h3 {
    font-family: "Arial";
    font-weight: bold;
    font-size: 20px;
  }

  :root {
    --primary-color: #00B4F0;
    --light-primary-color: #7CD8F7;
    --dark-primary-color: #0078A0;
    --error-color: #E61E6E;
  }

  h3 .anchor {
    color: grey;
    text-decoration: none;
    opacity: 0;
    font-size: 14px;
    margin-left: -1.6em;
    width: 1.6em;
    display: inline-block;
    outline: none;
  }

  h3:hover .anchor {
    opacity: 1;
  }

  #tabs ul {
    font-size: 20px;
    font-weight: bold;
    list-style: none;
    margin: 0;
    padding: 0;
    text-align: left;
  }

  #tabs li {
    background: #00ace6;
    border-radius: 15px 15px 0 0;
    display: inline;
    margin: 1px 1px 2px 0;
    padding: .75em 1em;
    padding-bottom: 4px;
    padding-top: 4px;
  }

  #tabs #nav li:hover {
    background: #0086b3;
  }

  #tabs #nav #selected li:hover {
    background: #00ace6;
  }

  #tabs a {
    color: white;
    padding: 0 1em;
    text-decoration: none;
  }

  #tabs a:hover {
    padding: 0 1em;
  }

  #tabs #nav a:hover {
    color: white;
  }

  #tabs #selected {
    border: none;
    background: #006080;
  }

  #tabs #selected a:hover {
    border: none;
    #background: #006080;
  }

  #tabs #selected a {
    color: white;
  }
